// Program 
// Application  : Event Driven Programming Sample
// Purpose      : Demonstrate a window with a non-modal read
// File Name    : DIALOG.PRG
// Section      :
// Category     :
// Description  : Demonstrate a window with a non-modal read
// Author       : Erik Wynn
// Copyright    : (c) 1993 by Wynn Systems Development Group, Ltd.
// Date created : 04-21-93              Date updated: 04-21-93
// Time created : 02:03:25pm            Time updated: 02:03:25pm
//

// Function/Procedure Prototype Table  -  Last Update: 04-21-93 @ 05:46:29pm
// 
// Return Value         Function/Arguments
//   
// Void                 PROCEDURE DialogProc( aMsg )
// Void                 PROCEDURE DlgInit( aWind )
// Void                 PROCEDURE DlgMove( aWind, nDir )
// aWind                FUNCTION wDialogNew( cTitle, nT, nL, nB, nR )

#include "msg.ch"
#include "wind.ch"
#include "inkey.ch"
#include "misc.ch"
#include "setcurs.ch"
#include "get.ch"
#include "getexit.ch"

FUNCTION wDialogNew( cTitle, nT, nL, nB, nR, bDialog )

	LOCAL aDialog := ARRAY( GET_ASIZE )
	LOCAL aWind

	aDialog[ GET_GETLIST ]   := {}
	aDialog[ GET_DIALOG ]    := bDialog
	aDialog[ GET_POS ]       := 1
	aDialog[ GET_LASTPOS ]   := 1
	aDialog[ GET_EXITSTATE ] := GE_NOEXIT
	aDialog[ GET_LASTEXIT ]  := GE_NOEXIT
	aDialog[ GET_BUMPTOP ]   := FALSE
	aDialog[ GET_BUMPBOT ]   := FALSE
	aDialog[ GET_UPDATED ]   := FALSE
	aDialog[ GET_KILLREAD ]  := FALSE

	aWind := wNew( cTitle, nT, nL, nB, nR, { |aMsg| DialogProc( aMsg ) }, aDialog )

	aDialog[ GET_WINDOW ]    := aWind

RETURN( aWind )


PROCEDURE DlgMove( aWind, nDir )

	LOCAL aGetList := aWind[ W_CARGO ][ GET_GETLIST ]
	LOCAL nGets := LEN( aGetList )
	LOCAL i
	LOCAL oGet

	// move the window
	IF wMove( aWind, nDir )
		// move the gets
		FOR i := 1 TO nGets
			oGet := aGetList[ i ]
			DO CASE
			CASE ( nDIR == K_CTRL_LEFT )  ; --oGet:col
			CASE ( nDIR == K_CTRL_RIGHT ) ; ++oGet:col
			CASE ( nDIR == K_CTRL_UP )    ; --oGet:row
			CASE ( nDIR == K_CTRL_DOWN )  ;	++oGet:row
			ENDCASE
		NEXT

	ENDIF

RETURN


PROCEDURE DlgGetFocus( aMsg )

	LOCAL aWind := aMsg[ M_WINDOW ]

	// restore focus for the current get
	ReadNonModal( aMsg )

	// get focus for the window
	wGetFocus( aWind )

RETURN


PROCEDURE DlgKillFocus( aMsg )

	LOCAL aWind := aMsg[ M_WINDOW ]

	// kill focus for the current get
	ReadNonModal( aMsg )

	// kill focus for the window
	wKillFocus( aWind )

RETURN


PROCEDURE DialogProc( aMsg )

	LOCAL aWind   := aMsg[ M_WINDOW ]
	LOCAL msg     := aMsg[ M_MSG ]
	LOCAL nParm   := aMsg[ M_PARM ]
	LOCAL nSavCrs

	DO CASE
	CASE ( msg == WM_CHAR )
		// pass the message down to ReadNonModal
		ReadNonModal( aMsg )
		SETCURSOR( nSavCrs )

	CASE ( msg == WM_CREATE )
		// different dialogs for different windows
		EVAL( aWind[ W_CARGO ][ GET_DIALOG ], aMsg )

	CASE ( msg == WM_PAINT )
		wDisp( aWind )

	CASE ( msg == WM_SETFOCUS )
		DlgGetFocus( aMsg )

	CASE ( msg == WM_KILLFOCUS )
		DlgKillFocus( aMsg )

	CASE ( msg == WM_MOVE )
		DlgMove( aWind, nParm )

	CASE ( msg == WM_HIDE )
		wHide( aWind )

	CASE ( msg == DLG_SAVE )
		// save data
		EVAL( aWind[ W_CARGO ][ GET_DIALOG ], aMsg )

	CASE ( msg == DLG_ABORT )
		// abort dialog
		EVAL( aWind[ W_CARGO ][ GET_DIALOG ], aMsg )

	CASE ( msg == WM_DESTROY )
		wKill( aWind )

	ENDCASE

	aWind[ W_LASTTIME ] := seconds()

RETURN


