// Program 
// Application  : Event Driven Programming Sample
// Purpose      : Demonstrate non-modal reads
// File Name    : NMREAD.PRG
// Section      :
// Category     :
// Description  : Demonstrate non-modal reads
// Author       : Erik Wynn
// Copyright    : (c) 1993 by Wynn Systems Development Group, Ltd.
// Date created : 04-21-93              Date updated: 04-21-93
// Time created : 07:09:47pm            Time updated: 07:09:47pm
//

#include "set.ch"
#include "inkey.ch"
#include "getexit.ch"
#include "wind.ch"
#include "msg.ch"
#include "misc.ch"
#include "get.ch"

#define K_UNDO K_CTRL_U

PROCEDURE ReadNonModal( aMsg )

	LOCAL aWind    := aMsg[ M_WINDOW ]
	LOCAL nKey     := aMsg[ M_PARM ]
	LOCAL msg      := aMsg[ M_MSG ]
	LOCAL aDialog  := aWind[ W_CARGO ]
	LOCAL aGetList := aDialog[ GET_GETLIST ]
	LOCAL oGet     := aGetList[ aDialog[ GET_POS ] ]
	LOCAL nPos
	LOCAL nExitState

	DO CASE
	CASE ( msg == WM_CHAR )

		// check for initial typeout (no editable positions)
		IF ( oGet:typeOut )
			oGet:exitState := GE_ENTER
		ENDIF

		IF ( VALTYPE( oGet:reader ) == "B" )
			// use user defined key handler
			EVAL( oGet:reader, oGet, nKey )
		ELSE
			// use the standard key handler
			GetApplyKey( oGet, nKey )
		ENDIF

		IF oGet:exitState != GE_NOEXIT
			// disallow exit if the VALID condition is not satisfied
			IF ( ! GetPostValidate( aDialog, oGet ) )
				oGet:exitState := GE_NOEXIT
			ENDIF
		ENDIF

		// settle, if required
		IF ( ( nExitState := oGet:exitState ) != GE_NOEXIT )
			oGet:killFocus()

			IF ( ( nPos := Settle( aDialog, aDialog[ GET_POS ] ) ) != 0 )
				// set next position in the list, and give focus to the get
				aGetList[ ( aDialog[ GET_POS ] := nPos ) ]:setFocus():exitState := GE_NOEXIT
			ELSE
				// end of the gets....
				IF nExitState == GE_ESCAPE
					// send the DLG_ABORT message back to the parent
					aMsg[ M_MSG ] := DLG_ABORT
					SendMessage( aMsg )

				ELSEIF nExitState == GE_WRITE
					// send the DLG_SAVE message back to the parent
					aMsg[ M_MSG ] := DLG_SAVE
					SendMessage( aMsg )

				ENDIF

			ENDIF

		ENDIF

	CASE ( ( msg == WM_KILLFOCUS ) .or. ( msg == WM_DESTROY ) )
		// kill focus for current get
		// oGet:KillFocus()

	CASE ( msg == WM_SETFOCUS )
		// set focus to oGet if the WHEN condition is satisfied
		IF ( GetPreValidate( aDialog, oGet ) )
			oGet:setFocus()
		ENDIF

	ENDCASE

RETURN


PROCEDURE GetApplyKey( oGet, nKey )

	LOCAL cKey

	DO CASE
	CASE ( nKey == K_UP )
		oGet:exitState := GE_UP

	CASE ( nKey == K_SH_TAB )
		oGet:exitState := GE_UP

	CASE ( nKey == K_DOWN )
		oGet:exitState := GE_DOWN

	CASE ( nKey == K_TAB )
		oGet:exitState := GE_DOWN

	CASE ( nKey == K_ENTER )
		oGet:exitState := GE_ENTER

	CASE ( nKey == K_ESC )
		IF ( Set( _SET_ESCAPE ) )
			oGet:undo()
			oGet:exitState := GE_ESCAPE
		END

	CASE ( nKey == K_PGUP )
		oGet:exitState := GE_WRITE

	CASE ( nKey == K_PGDN )
		oGet:exitState := GE_WRITE

	CASE ( nKey == K_CTRL_HOME )
		oGet:exitState := GE_TOP

#ifdef CTRL_END_SPECIAL

	// both ^W and ^End go to the last GET
	CASE ( nKey == K_CTRL_END )
		oGet:exitState := GE_BOTTOM

#else

	// both ^W and ^End terminate the READ (the default)
	CASE ( nKey == K_CTRL_W )
		oGet:exitState := GE_WRITE

#endif

	CASE ( nKey == K_INS )
		Set( _SET_INSERT, !Set( _SET_INSERT ) )

		// yeah, right...
		// ShowScoreboard()

	CASE ( nKey == K_UNDO )
		oGet:Undo()

	CASE ( nKey == K_HOME )
		oGet:Home()

	CASE ( nKey == K_END )
		oGet:End()

	CASE ( nKey == K_RIGHT )
		oGet:Right()

	CASE ( nKey == K_LEFT )
		oGet:Left()

	CASE ( nKey == K_CTRL_RIGHT )
		oGet:WordRight()

	CASE ( nKey == K_CTRL_LEFT )
		oGet:WordLeft()

	CASE ( nKey == K_BS )
		oGet:BackSpace()

	CASE ( nKey == K_DEL )
		oGet:Delete()

	CASE ( nKey == K_CTRL_T )
		oGet:DelWordRight()

	CASE ( nKey == K_CTRL_Y )
		oGet:DelEnd()

	CASE ( nKey == K_CTRL_BS )
		oGet:DelWordLeft()

	OTHERWISE

		IF ( nKey >= 32 .and. nKey <= 255 )

			cKey := CHR( nKey )

			IF ( oGet:type == "N" .and. ( cKey == "." .or. cKey == "," ) )
				oGet:ToDecPos()

			ELSE
				IF ( Set( _SET_INSERT ) )
					oGet:Insert( cKey )
				ELSE
					oGet:Overstrike( cKey )
				END

				IF ( oGet:typeOut .and. ! Set( _SET_CONFIRM ) )
					IF ( Set( _SET_BELL ) )
						?? CHR( 7 )
					ENDIF

					oGet:exitState := GE_ENTER
				ENDIF

			ENDIF

		ENDIF

	ENDCASE

RETURN


STATIC FUNCTION Settle( aDialog, nPos )

	LOCAL aGetList := aDialog[ GET_GETLIST ]
	LOCAL exitState

	IF ( nPos == 0 )
		exitState := GE_DOWN
	ELSE
		exitState := aGetList[ nPos ]:exitState
	ENDIF

	IF ( exitState == GE_ESCAPE .or. exitState == GE_WRITE )
		RETURN( 0 )
	ENDIF

	IF ( exitState <> GE_WHEN )
		// reset state info
		aDialog[ GET_LASTPOS ] := nPos

	ELSE
		// re-use last exitState, do not disturb state info
		exitState := aDialog[ GET_LASTEXIT ]

	ENDIF

	DO CASE
	CASE ( exitState == GE_UP )
		nPos--

	CASE ( exitState == GE_DOWN )
		nPos++

	CASE ( exitState == GE_TOP )
		nPos := 1
		aDialog[ GET_BUMPTOP ] := TRUE
		exitState := GE_DOWN

	CASE ( exitState == GE_BOTTOM )
		nPos := LEN( aGetList )
		aDialog[ GET_BUMPTOP ] := TRUE
		exitState := GE_UP

	CASE ( exitState == GE_ENTER )
		nPos++

	ENDCASE

	IF ( nPos == 0 )

		IF ( ! ReadExit() .and. ! aDialog[ GET_BUMPTOP ] )
			aDialog[ GET_BUMPTOP ] := TRUE
			nPos := aDialog[ GET_LASTPOS ]
			exitState := GE_DOWN
		ENDIF

	ELSEIF ( nPos == LEN( aGetList ) + 1 )

		IF ( ! ReadExit() .and. exitState <> GE_ENTER .and. ! aDialog[ GET_BUMPTOP ] )
			aDialog[ GET_BUMPBOT ] := TRUE
			nPos := aDialog[ GET_LASTPOS ]
			exitState := GE_UP
		ELSE
			nPos := 0
		ENDIF

	ENDIF

	// record exit state
	aDialog[ GET_LASTEXIT ] := exitState

	IF ( nPos <> 0 )
		aGetList[ nPos ]:exitState := exitState
	ENDIF

RETURN( nPos )


FUNCTION ReadExit( lNew )
RETURN( SET( _SET_EXIT, lNew ) )


FUNCTION GetPreValidate( aDialog, oGet )

	LOCAL saveUpdated
	LOCAL lWhen := TRUE

	IF ( oGet:preBlock <> NIL )

		saveUpdated := aDialog[ GET_UPDATED ]

		lWhen := Eval( oGet:preBlock, oGet )

		oGet:Display()

		// yeah, right...
		// ShowScoreBoard()
		aDialog[ GET_UPDATED ] := saveUpdated

	ENDIF


	IF ( aDialog[ GET_KILLREAD ] )
		lWhen := FALSE
		// provokes ReadModal() exit
		oGet:exitState := GE_ESCAPE

	ELSEIF ( !lWhen )
		// indicates failure
		oGet:exitState := GE_WHEN

	ELSE
		// prepares for editing
		oGet:exitState := GE_NOEXIT

	ENDIF

RETURN( lWhen )


FUNCTION GetPostValidate( aDialog, oGet )

	LOCAL saveUpdated
	LOCAL changed
	LOCAL valid := TRUE

	IF ( oGet:exitState == GE_ESCAPE )
		RETURN( TRUE )
	ENDIF

	IF ( oGet:BadDate() )
		oGet:Home()
		// yeah, right
		// DateMsg()
		// ShowScoreboard()

		RETURN( FALSE )
	ENDIF

	// if editing occurred, assign the new value to the variable
	IF ( oGet:changed )
		oGet:Assign()
		aDialog[ GET_UPDATED ] := TRUE

	ENDIF

	// reform edit buffer, set cursor to home position, redisplay
	oGet:Reset()

	// check VALID condition if specified
	IF ( oGet:postBlock <> NIL )

		saveUpdated := aDialog[ GET_UPDATED ]

		// S87 compat.
		SetPos( oGet:row, oGet:col + LEN( oGet:buffer ) )

		valid := EVAL( oGet:postBlock, oGet )

		// reset compat. pos
		SetPos( oGet:row, oGet:col )

		// yeah, right...
		// ShowScoreBoard()

		oGet:UpdateBuffer()

		aDialog[ GET_UPDATED ] := saveUpdated

		IF ( aDialog[ GET_KILLREAD ] )
			// provokes ReadModal() exit
			oGet:exitState := GE_ESCAPE
			valid := TRUE
		ENDIF

	ENDIF

RETURN( valid )

