//  Program Ŀ
//  Application  : Event Driven Programming Sample                           
//  Description  : Window Functions                                          
//  File Name    : WINDOW.PRG                                                
//  Author       : Erik Wynn                                                 
//  Date created : 02-02-93              Date updated: 02-02-93             
//  Time created : 01:52:18pm            Time updated: 01:52:18pm           
//  Copyright    : (c) 1993 by Wynn Systems Development Group, Ltd.          
// 

// Function/Procedure Prototype Table  -  Last Update: 04-21-93 @ 04:59:10pm
// 
// Return Value         Function/Arguments
//   
// Void                 PROCEDURE KillTimer( aWind )
// Void                 PROCEDURE SendMessage( aWind, aMsg )
// Void                 PROCEDURE SetTimer( aWind, nTimer )
// Void                 PROCEDURE wAdd( aWind )
// Void                 PROCEDURE wCsrAdvance( aWind )
// Void                 PROCEDURE wCsrRewind( aWind )
// Void                 PROCEDURE wDelete( aWind )
// Void                 PROCEDURE wDisp( aWind )
// Void                 PROCEDURE wGetFocus( aWind )
// Void                 PROCEDURE wHide( aWind )
// Void                 PROCEDURE wKill( aWind )
// Void                 PROCEDURE wKillFocus( aWind )
// Void                 PROCEDURE wLimit( aWind, nT, nL, nB, nR )
// Void                 PROCEDURE wLineAdvance( aWind )
// Void                 PROCEDURE wPrintChar( aWind, cChar )
// Void                 PROCEDURE wPrintText( aWind, nChar )
// Void                 PROCEDURE wSay( aWind, nRow, nCol, cText, cColor )
// Void                 PROCEDURE wScroll( aWind, t, l, b, r )
// Void                 PROCEDURE wTitleDisp( aWind, lFocus )
// saCurrent            FUNCTION SetFocus( aWind )
// saCurrent            FUNCTION wCurrent()
// saFirst              FUNCTION wFirst()
// saLast               FUNCTION wLast()
// lMoved               FUNCTION wMove( aWind, nDir )
// aWind                FUNCTION wNew( cTitle, nT, nL, nB, nR, bProc, xCargo )
// IIF( aWind[ W_NE...  FUNCTION wNext( aWind )
// IIF( aWind[ W_PR...  FUNCTION wPrev( aWind )


#include "setcurs.ch"
#include "inkey.ch"
#include "wind.ch"
#include "msg.ch"
#include "misc.ch"

STATIC saFirst                                   // first window in list
STATIC saLast                                    // last window in list
STATIC saCurrent                                 // current window


FUNCTION wNew( cTitle, nT, nL, nB, nR, bProc, xCargo )

	LOCAL aWind := ARRAY( W_ASIZE )
	LOCAL aMsg  := ARRAY( M_ASIZE )

	//  create the window pseudo-object
	aWind[ W_TITLE ]     := cTitle
	aWind[ W_T ]         := nT
	aWind[ W_L ]         := nL
	aWind[ W_B ]         := nB
	aWind[ W_R ]         := nR
	aWind[ W_CURS_ROW ]  := 1                       // row relative to top
	aWind[ W_CURS_COL ]  := 1                       // col relative to left
	aWind[ W_PROC ]      := bProc
	aWind[ W_CARGO ]     := xCargo
	aWind[ W_TIMER ]     := NIL
	aWind[ W_LASTTIME ]  := 0
	aWind[ W_PREVSCR ]   := SAVESCREEN( nT, nL, nB, nR )
	aWind[ W_CURRSCR ]   := NIL
	aWind[ W_HASFOCUS ]  := TRUE
	aWind[ W_AREA_CLR ]  := CLR_WIND               // default
	aWind[ W_TTL_CLR ]   := CLR_TITLE              // default
	aWind[ W_FOCUS_CLR ] := CLR_FOCUS              // default
	aWind[ W_TEXT_CLR ]  := CLR_TEXT               // default
	aWind[ W_GET_CLR ]   := CLR_GET                // default

	aWind[ W_T_LIMIT ]   := 0
	aWind[ W_L_LIMIT ]   := 0
	aWind[ W_B_LIMIT ]   := MAXROW()
	aWind[ W_R_LIMIT ]   := MAXCOL()

	aWind[ W_HIDDEN ]    := TRUE

	wAdd( aWind )

	// send the WM_CREATE message to the window
	aMsg[ M_WINDOW ] := aWind
	aMsg[ M_MSG ] := WM_CREATE
	SendMessage( aMsg )

RETURN( aWind )


PROCEDURE wAdd( aWind )

	// check for new saLast
	IF saLast != NIL
		saLast[ W_NEXT ] := aWind
	ENDIF

	// check for new saFirst
	IF ( saFirst == NIL )
		saFirst := aWind
	ENDIF

	// kill focus of current window
	IF ( saCurrent != NIL )
		saCurrent[ W_HASFOCUS ] := FALSE
	ENDIF

	// set prev and next values
	aWind[ W_PREV ] := saLast
	aWind[ W_NEXT ] := NIL

	// make this one current
	saLast    := aWind
	saCurrent := aWind
	aWind[ W_HASFOCUS ] := TRUE

RETURN


PROCEDURE wDelete( aWind )

		LOCAL aPrev  := aWind[ W_PREV ]
		LOCAL aNext  := aWind[ W_NEXT ]
		LOCAL aLast  := saLast
		LOCAL aFirst := saFirst

		// first, reset prev and next window next and prev pointers
		IF ( aPrev != NIL )
			aPrev[ W_NEXT ] := aNext
		ENDIF

		IF ( aNext != NIL )
			aNext[ W_PREV ] := aPrev
		ENDIF

		// now, check if the window was first or last
		IF ( aLast  == aWind )
			saLast  := aPrev
		ENDIF

		IF ( aFirst == aWind )
			saFirst := aNext
		ENDIF

RETURN


PROCEDURE wKill( aWind )

	LOCAL aNext := wNext()

	// remove window from screen
	wHide( aWind )

	// remove window from list
	wDelete( aWind )

	// set current window to NIL
	saCurrent := NIL

	// set focus to the next window in the list
	setFocus( aNext )

RETURN



PROCEDURE wLimit( aWind, nT, nL, nB, nR )

	IF nT == NIL
		nT := 0
	ENDIF

	IF nL == NIL
		nL := 0
	ENDIF

	IF nB == NIL
		nB := MAXROW()
	ENDIF

	IF nR == NIL
		nR := MAXCOL()
	ENDIF

	aWind[ W_T_LIMIT ] := nT
	aWind[ W_L_LIMIT ] := nL
	aWind[ W_B_LIMIT ] := nB
	aWind[ W_R_LIMIT ] := nR

RETURN


PROCEDURE wDisp( aWind )

	LOCAL cOldColor

	DISPBEGIN()

	// window is no longer hidden
	aWind[ W_HIDDEN ] := FALSE

	aWind[ W_PREVSCR ] := SAVESCREEN( aWind[ W_T ], aWind[ W_L ], aWind[ W_B ], aWind[ W_R ] )
	IF ( aWind[ W_CURRSCR ] != NIL )
		RESTSCREEN( aWind[ W_T ], aWind[ W_L ], aWind[ W_B ], aWind[ W_R ], aWind[ W_CURRSCR ] )
	ELSE
		cOldColor := SETCOLOR( aWind[ W_AREA_CLR ] )
		SCROLL( aWind[ W_T ], aWind[ W_L ], aWind[ W_B ], aWind[ W_R ], 0 )
		SETCOLOR( cOldColor )
	ENDIF

	// display title
	wTitleDisp( aWind )


	DISPEND()

RETURN


PROCEDURE wHide( aWind )

	// save current window image
	aWind[ W_CURRSCR ] := SAVESCREEN( aWind[ W_T ], aWind[ W_L ], aWind[ W_B ], aWind[ W_R ] )

	// restore image under window
	RESTSCREEN( aWind[ W_T ], aWind[ W_L ], aWind[ W_B ], aWind[ W_R ], aWind[ W_PREVSCR ] )
	aWind[ W_HIDDEN ] := TRUE

RETURN


PROCEDURE wTitleDisp( aWind, lFocus )

	IF lFocus == NIL
		lFocus := aWind[ W_HASFOCUS ]
	ENDIF

	wSay( aWind, 0, 0, PADC( aWind[ W_TITLE ], aWind[ W_R ] - aWind[ W_L ] + 1 ), ;
		IIF( lFocus, aWind[ W_FOCUS_CLR ], aWind[ W_TTL_CLR ] ) )

RETURN


PROCEDURE wSay( aWind, nRow, nCol, cText, cColor )

	LOCAL nOldRow := ROW()
	LOCAL nOldCol := COL()
	LOCAL aMsg    := ARRAY( M_ASIZE )
	LOCAL aCurrWindow := aWind
	LOCAL aWindows := {}
	LOCAL nWindows
	LOCAL i
	LOCAL lHidden := aWind[ W_HIDDEN ]

	// take care of default parameters
	IF ( cText == NIL )
		cText := ""
	ENDIF

	IF ( cColor == NIL )
		cColor := aWind[ W_TEXT_CLR ]
	ENDIF

	// begin video buffering
	DISPBEGIN()

	IF ( ! aWind[ W_HASFOCUS ] )

		aCurrWindow := saLast
		// hide all (visible) greater windows in the list, starting with the last one
		WHILE ( ! aCurrWindow == aWind )
			IF ! aCurrWindow[ W_HIDDEN ]
				wHide( aCurrWindow )
				AADD( aWindows, aCurrWindow )
			ENDIF
			aCurrWindow := aCurrWindow[ W_PREV ]
		END

		// remember the next window in the list
		aCurrWindow := aCurrWindow[ W_NEXT ]

	ENDIF

	// check to see if this window is hidden
	IF ( lHidden )
		// display window for writing purposes
		wDisp( aWind )
	ENDIF

	// display text
	DEVPOS( aWind[ W_T ] + nRow, aWind[ W_L ] + nCol )
	DEVOUT( cText, cColor )

	// save screen image
	aWind[ W_CURRSCR ] := SAVESCREEN( aWind[ W_T ], aWind[ W_L ], aWind[ W_B ], aWind[ W_R ] )

	IF ( lHidden )
		// hide window again -- the fast way
		RESTSCREEN( aWind[ W_T ], aWind[ W_L ], aWind[ W_B ], aWind[ W_R ], aWind[ W_PREVSCR ] )
		aWind[ W_HIDDEN ] := TRUE
	ENDIF

	IF ( ! aWind[ W_HASFOCUS ] )

		// redisplay all windows in saved list
		nWindows := LEN( aWindows )
		FOR i := nWindows TO 1 STEP -1
			aMsg[ M_WINDOW ] := aWindows[ i ]
			aMsg[ M_MSG ] := WM_PAINT
			SendMessage( aMsg )
		NEXT

	ENDIF

	// restore cursor position
	DEVPOS( nOldRow, nOldCol )

	// end video buffering
	DISPEND()

RETURN


PROCEDURE wScroll( aWind, t, l, b, r )

	LOCAL cOldColor := SETCOLOR( aWind[ W_AREA_CLR ] )

	SCROLL( t, l, b, r )
	aWind[ W_CURRSCR ] := SAVESCREEN( aWind[ W_T ], aWind[ W_L ], aWind[ W_B ], aWind[ W_R ] )

	SETCOLOR( cOldColor )

RETURN


FUNCTION wMove( aWind, nDir )

	LOCAL aMsg := ARRAY( M_ASIZE )
	LOCAL aCurrWindow
	LOCAL lMoved := FALSE

	// start video buffering
	DISPBEGIN()

	// save contents of window
	aWind[ W_CURRSCR ] := SAVESCREEN( aWind[ W_T ], aWind[ W_L ], aWind[ W_B ], aWind[ W_R ] )

	// restore image under window
	RESTSCREEN( aWind[ W_T ], aWind[ W_L ], aWind[ W_B ], aWind[ W_R ], aWind[ W_PREVSCR ] )

	// move window
	DO CASE
	CASE ( nDIR == K_CTRL_LEFT )
		IF ( aWind[ W_L ] > aWind[ W_L_LIMIT ] )
			--aWind[ W_L ]
			--aWind[ W_R ]
			lMoved := TRUE
		ENDIF

	CASE ( nDIR == K_CTRL_RIGHT )
		IF ( aWind[ W_R ] < aWind[ W_R_LIMIT ] )
			++aWind[ W_L ]
			++aWind[ W_R ]
			lMoved := TRUE
		ENDIF

	CASE ( nDIR == K_CTRL_UP )
		IF ( aWind[ W_T ] > aWind[ W_T_LIMIT ] )
			--aWind[ W_T ]
			--aWind[ W_B ]
			lMoved := TRUE
		ENDIF

	CASE ( nDIR == K_CTRL_DOWN )
		IF ( aWind[ W_B ] < aWind[ W_B_LIMIT ] )
			++aWind[ W_T ]
			++aWind[ W_B ]
			lMoved := TRUE
		ENDIF

	ENDCASE

	// re-display window
	aMsg[ M_WINDOW ] := aWind
	aMsg[ M_MSG ]    := WM_PAINT
	SendMessage( aMsg )

	// end video buffering
	DISPEND()

RETURN( lMoved )



FUNCTION wCurrent()
RETURN( saCurrent )


PROCEDURE wGetFocus( aWind )

	wTitleDisp( aWind, TRUE )

RETURN


PROCEDURE wKillFocus( aWind )

	wTitleDisp( aWind, FALSE )

RETURN


FUNCTION SetFocus( aWind )

	LOCAL aMsg := ARRAY( M_ASIZE )
	LOCAL aCurrWindow
	LOCAL aWindows := {}
	LOCAL nWindows
	LOCAL i
	LOCAL lHidden := aWind[ W_HIDDEN ]

	IF ( ! aWind[ W_HASFOCUS ] )

		// start video buffering
		DISPBEGIN()

		IF saCurrent != NIL
			// save cursor status
			saCurrent[ W_CURS_STAT ] := SETCURSOR()

			// send the kill focus message to the current window
			aMsg[ M_WINDOW ] := saCurrent
			aMsg[ M_MSG ]    := WM_KILLFOCUS
			SendMessage( aMsg )

			// remove focus from current window
			saCurrent[ W_HASFOCUS ] := FALSE
		ENDIF

		aCurrWindow := saLast
		// hide all (visible) greater windows in the list, starting with the last one
		WHILE ( ! aCurrWindow == aWind )
			IF ! aCurrWindow[ W_HIDDEN ]
				wHide( aCurrWindow )
				AADD( aWindows, aCurrWindow )
			ENDIF
			aCurrWindow := aCurrWindow[ W_PREV ]
		END

		// hide this one too, if it is not already hidden
		IF ! lHidden
			wHide( aWind )
		ENDIF

		// remember the next window in the list
		aCurrWindow := aCurrWindow[ W_NEXT ]

		// make this the current window -- remove, then add
		wDelete( aWind )
		wAdd( aWind )

		// re-display all windows in the saved list from marker
		// -- check for beginning of list marker
		IF aCurrWindow == NIL
			aCurrWindow := saFirst
		ENDIF

		nWindows := LEN( aWindows )
		FOR i := nWindows TO 1 STEP -1
			aMsg[ M_WINDOW ] := aWindows[ i ]
			aMsg[ M_MSG ] := WM_PAINT
			SendMessage( aMsg )
		NEXT

		// if not hidden, redisplay this window
		IF ! lHidden
			aMsg[ M_WINDOW ] := aWind
			aMsg[ M_MSG ] := WM_PAINT
			SendMessage( aMsg )

			// restore cursor status
			SETCURSOR( aWind[ W_CURS_STAT ] )

		ENDIF

		// send the get focus message to the window
		aMsg[ M_WINDOW ] := aWind
		aMsg[ M_MSG ]    := WM_SETFOCUS
		SendMessage( aMsg )

		// end video buffering
		DISPEND()

	ENDIF

RETURN( saCurrent )


FUNCTION wNext( aWind )

	IF ( aWind == NIL )
		aWind := saCurrent
	ENDIF

RETURN( IIF( aWind[ W_NEXT ] == NIL, saFirst, aWind[ W_NEXT ] ) )


FUNCTION wPrev( aWind )

	IF ( aWind == NIL )
		aWind := saCurrent
	ENDIF

RETURN( IIF( aWind[ W_PREV ] == NIL, saLast, aWind[ W_PREV ] ) )


PROCEDURE wPrintText( aWind, nChar )

	wPrintChar( aWind, CHR( nChar ) )
	wCsrAdvance( aWind )

RETURN


PROCEDURE wPrintChar( aWind, cChar )

	wSay( aWind, aWind[ W_CURS_ROW ], aWind[ W_CURS_COL ], cChar, aWind[ W_TEXT_CLR ] )

RETURN


PROCEDURE wCsrAdvance( aWind )

	IF ( ++aWind[ W_CURS_COL ] == aWind[ W_R ] - aWind[ W_L ] )
		aWind[ W_CURS_COL ] := 1
		wLineAdvance( aWind )
	ELSE
		// reposition cursor
		DEVPOS( aWind[ W_T ] + aWind[ W_CURS_ROW ], aWind[ W_L ] + aWind[ W_CURS_COL ] )
	ENDIF

RETURN


PROCEDURE wLineAdvance( aWind )

	LOCAL cOldColor := SETCOLOR( aWind[ W_AREA_CLR ] )

	IF ( aWind[ W_CURS_ROW ]  == aWind[ W_B ] - aWind[ W_T ] - 1 )
		SCROLL( aWind[ W_T ] + 1, aWind[ W_L ] + 1, aWind[ W_B ] - 1, aWind[ W_R ] - 1, 1 )
	ELSE
		aWind[ W_CURS_ROW ]++
	ENDIF

	aWind[ W_CURS_COL ] := 1

	// reposition cursor
	DEVPOS( aWind[ W_T ] + aWind[ W_CURS_ROW ], aWind[ W_L ] + aWind[ W_CURS_COL ] )
	SETCOLOR( cOldColor )

RETURN


PROCEDURE wCsrRewind( aWind )

	IF --aWind[ W_CURS_COL ] == 0
		aWind[ W_CURS_ROW ]--
		aWind[ W_CURS_COL ] := aWind[ W_R ] - 1
	ENDIF

	// reposition cursor
	DEVPOS( aWind[ W_T ] + aWind[ W_CURS_ROW ], aWind[ W_L ] + aWind[ W_CURS_COL ] )

RETURN


FUNCTION wFirst()
RETURN( saFirst )


FUNCTION wLast()
RETURN( saLast )


PROCEDURE SendMessage( aMsg )

	EVAL( aMsg[ M_WINDOW ][ W_PROC ], aMsg )

RETURN


PROCEDURE SetTimer( aWind, nTimer )

	aWind[ W_TIMER ] := nTimer

RETURN


PROCEDURE KillTimer( aWind )

	aWind[ W_TIMER ] := NIL

RETURN

// EOF: Window.prg
